package nl.logicacmg.grid.client;
import java.awt.BorderLayout;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.StringTokenizer;

import javax.swing.ImageIcon;
import javax.swing.JFrame;

import nl.logicacmg.grid.service.mandelbrot.MandelbrotReply;
import nl.logicacmg.grid.service.mandelbrot.MandelbrotRequest;

import org.apache.log4j.Logger;

/**
 * Testclient for the Fractal Server <br>
 * 
 * Originally (20-08-2003) written for the 'hintme' image server <br>
 * Now used for current project 'Fractal Grid Service' <br>
 * @author Wico Mulder, 10-08-2004
 * 
 * The following keys are used to control this client:<br>
 * up       - move fractal area up <br> 
 * down     - move fractal area down <br>
 * left     - move fractal area left <br>
 * right    - move fractal area right <br>
 * x/X      - increase/decrease width (no of pixels) <br>
 * y/Y      - increase/decrease height (no of pixels) <br>
 * space    - flip between real size/stretched view <br>
 * pageup   - zoom in <br>
 * pagedown - zoom out <br>
 * i/I      - increase/decrease no of iterations that is used <br>
 *            for calculating one pixel of the fractal <br>
 * s        - reset to standard view
 * l 		- direct large calculation request (1024x768)
 * p        - save to file
 * r		- reload property file
 */

public class MandelbrotViewer extends JFrame implements KeyListener, Runnable {
	private ImagePanel imagepanel;
	private InfoPanel infopanel;
	private double minx,oldminx;
	private double miny,oldminy;
	private double maxx,oldmaxx;
	private double maxy,oldmaxy;
	private int width;
	private int height;
	private int iterations;
	private boolean up;
	private boolean down;
	private boolean left;
	private boolean right;
	private boolean zoomin;
	private boolean zoomout;
	private boolean stretch;
	private boolean inciterations;
	private boolean deciterations;
	private boolean incwidth;
	private boolean decwidth;
	private boolean incheight;
	private boolean decheight;
	private boolean makecall=true;
	private Thread client_thread;
    private MandelbrotBroker broker;
    private int[][] colorpeaks; 
	Logger logger=Logger.getLogger(MandelbrotBroker.class);
	
	public MandelbrotViewer() {
		broker=new MandelbrotBroker();
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setTitle("Fancy Broker Client");
		imagepanel = new ImagePanel();
		infopanel = new InfoPanel();
		getContentPane().add(imagepanel, BorderLayout.CENTER);
		getContentPane().add(infopanel, BorderLayout.NORTH);
		addKeyListener(this);
		requestFocus();
        init();
 		refresh();
		setSize(width+50, height+100);
		show();
		client_thread = new Thread(this);
		client_thread .start();
	}
    
    public void init(){
		minx = -2;
		miny = -1;
		maxx = 0.5;
		maxy = 1;
		width = PropertyManager.getViewerWidth();
		height = PropertyManager.getViewerHeight();
		iterations = 20;
		initColorPeaks();
    }
    
    public void initColorPeaks(){
        String peaks=PropertyManager.getColorPeaks();
        int count=0;
        for (int i=0;i<peaks.length();i++){
            if (peaks.charAt(i)==';') count++;
        }
        logger.debug("***"+count);
        colorpeaks = new int[count+1][3];
        count=0;
        StringTokenizer peaktokenizer=new StringTokenizer(peaks,";");
        while (peaktokenizer.hasMoreTokens()){
            String rgb=peaktokenizer.nextToken();
            logger.debug("***"+rgb);
                StringTokenizer rgbtokenizer=new StringTokenizer(rgb,",");
            colorpeaks[count][0]=Integer.parseInt(rgbtokenizer.nextToken());
            colorpeaks[count][1]=Integer.parseInt(rgbtokenizer.nextToken());
            colorpeaks[count][2]=Integer.parseInt(rgbtokenizer.nextToken());
            count++;
        }
				
    }

	/**
	 *  shifts the rectangular area [minx,miny,maxx,maxy] 
	 */
	private void delta(double factorx, double factory) {
		minx += factorx * (maxx - minx);
		maxx += factorx * (maxx - minx);
		miny += factory * (maxy - miny);
		maxy += factory * (maxy - miny);
	}

	/**
	 * zooms in/out the recangular area [minx,miny,maxx,maxy]
	 * @param factor - the zoom factor (usually 0.1)
	 */
	private void zoom(double factor) {
		double centerx = (maxx + minx) / 2.0;
		double centery = (maxy + miny) / 2.0;
		minx = centerx - ((centerx - minx) * (1.0 - factor));
		maxx = centerx + ((maxx - centerx) * (1.0 - factor));
		miny = centery - ((centery - miny) * (1.0 - factor));
		maxy = centery + ((maxy - centery) * (1.0 - factor));
	}

	public void keyPressed(KeyEvent evt) {
		if (evt.getKeyCode() == KeyEvent.VK_UP) {
			up = true;
		}
		if (evt.getKeyCode() == KeyEvent.VK_DOWN) {
			down = true;
		}
		if (evt.getKeyCode() == KeyEvent.VK_RIGHT) {
			right = true;
		}
		if (evt.getKeyCode() == KeyEvent.VK_LEFT) {
			left = true;
		}
		if (evt.getKeyCode() == KeyEvent.VK_PAGE_UP) {
			zoomin = true;
		}
		if (evt.getKeyCode() == KeyEvent.VK_PAGE_DOWN) {
			zoomout = true;
		}

		

		if (evt.getKeyChar() == 'i') {
			inciterations = true;
		}
		if (evt.getKeyChar() == 'I') {
			deciterations = true;
		}
		if (evt.getKeyChar() == 'x') {
			incwidth = true;
		}
		if (evt.getKeyChar() == 'X') {
			decwidth = true;
		}
		if (evt.getKeyChar() == 'y') {
			incheight = true;
		}
		if (evt.getKeyChar() == 'Y') {
			decheight = true;
		}

		if (evt.getKeyCode() == KeyEvent.VK_ALT) {
			makecall = false;
		}

	}

	public void keyTyped(KeyEvent evt) {
		if (evt.getKeyChar() == ' ') {
			stretch = !stretch;
			refresh();
		}
		if (evt.getKeyChar() == 's') {
			init();
			refresh();
		}
		
		if (evt.getKeyChar() == 'l') {
			width=PropertyManager.getHugeCalcWitdh();
					height = PropertyManager.getHugeCalcHeight();
					iterations = PropertyManager.getHugeCalcInterations();
					minx=PropertyManager.getHugeCalcMinx();;
					miny=PropertyManager.getHugeCalcMiny();;
					maxx=PropertyManager.getHugeCalcMaxx();;
					maxy=PropertyManager.getHugeCalcMaxy();;
			this.setSize(width,height);
            refresh();
		}
		
		if (evt.getKeyChar()=='r'){
		    PropertyManager.init();
		    initColorPeaks();
			refresh();
		}
		
		
		if (evt.getKeyChar() =='p'){
			PngEncoder encoder=new PngEncoder(imagepanel.getImage());
			byte[] bytes=encoder.pngEncode();
			try {
				System.out.println("write");
				FileOutputStream out=new FileOutputStream("out.png");
				BufferedOutputStream bout=new BufferedOutputStream(out);
				bout.write(bytes);
				bout.close();
				out.close();
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public void keyReleased(KeyEvent evt) {
		if (evt.getKeyCode() == KeyEvent.VK_UP) {
			up = false;
		}
		if (evt.getKeyCode() == KeyEvent.VK_LEFT) {
			left = false;
		}
		if (evt.getKeyCode() == KeyEvent.VK_RIGHT) {
			right = false;
		}
		if (evt.getKeyCode() == KeyEvent.VK_DOWN) {
			down = false;
		}
		if (evt.getKeyCode() == KeyEvent.VK_PAGE_UP) {
			zoomin = false;
		}
		if (evt.getKeyCode() == KeyEvent.VK_PAGE_DOWN) {
			zoomout = false;
		}

		if (evt.getKeyChar() == 'i') {
			inciterations = false;
		}
		if (evt.getKeyChar() == 'I') {
			deciterations = false;
		}
		if (evt.getKeyChar() == 'x') {
			incwidth = false;
		}
		if (evt.getKeyChar() == 'X') {
			decwidth = false;
		}
		if (evt.getKeyChar() == 'y') {
			incheight = false;
		}
		if (evt.getKeyChar() == 'Y') {
			decheight = false;
		}

		if (evt.getKeyCode() == KeyEvent.VK_ALT) {
			makecall = true;
			refresh();
		}

	}

	/**
	 * refreshes the image, by sending a request to the service <br>
	 * decode the returned String <br>
	 * and draw the image <br>
	 */
	public void refresh() {
		long t0 = System.currentTimeMillis();
		
		if (makecall){
			logger.debug("Broker  makes request for ["+minx+","+miny+","+maxx+","+maxy+"] "+width+"x"+height);
				MandelbrotRequest request =
				new MandelbrotRequest(
					width,
					height,
					minx,
					miny,
					maxx,
					maxy,
					iterations,colorpeaks);
			
			MandelbrotReply mandelbrotreply = (MandelbrotReply) broker.broke(request);
	    	
		ImageIcon i = new ImageIcon(mandelbrotreply.getImage());
		imagepanel.setImage(i.getImage(), stretch);
		oldminx=minx;
		oldminy=miny;
		oldmaxx=maxx;
		oldmaxy=maxy;
		} else {
			
			double a=(double)imagepanel.getWidth()/(oldmaxx-oldminx);
			double b=oldminx*-a;
			int x1=(int)(a*minx+b);
			int x2=(int)(a*maxx+b);
			a=(double)imagepanel.getHeight()/(oldmaxy-oldminy);
			b=oldminy*-a;
			int y1=(int)(a*miny+b);
			int y2=(int)(a*maxy+b);
			
			imagepanel.drawRect(x1,y1,x2-x1,y2-y1);
		}
		long t1 = System.currentTimeMillis();

		// build info strings
		infopanel.setText(
			" ["
				+ round(minx, 8)
				+ ","
				+ round(miny, 8)
				+ ","
				+ round(maxx, 8)
				+ ","
				+ round(maxy, 8)
				+ "]");
		infopanel.setText2(
			" i"
				+ iterations
				+ " "
				+ width
				+ "x"
				+ height
				+ "="
				+ width * height
				+ " "
				+ (t1 - t0)
				+ "ms");
	}

	/**
	 * utility method, rounds doubles for String (show) purpose
	 */
	private String round(double value, double noOfDecimals) {
		double tenths = Math.pow(10, noOfDecimals);
		double roundedDouble = Math.round(value * tenths) / tenths;
		return new Double(roundedDouble).toString();
	}

	/**
	 * Running thread, that actually waits for keypresses
	 * and checks for actions
	 */
	public void run() {
		while (true) {
			try {
				Thread.sleep(100);
			} catch (InterruptedException ie) {

			}
			// the actions.
			// use refresh flag, only call and draw if necessary
			boolean refresh = false;
			if (down) {
				delta(0, 0.1);
				refresh = true;
			}
			if (up) {
				delta(0, -0.1);
				refresh = true;
			}
			if (right) {
				delta(0.1, 0);
				refresh = true;
			}
			if (left) {
				delta(-0.1, 0);
				refresh = true;
			}
			if (zoomin) {
				zoom(0.1);
				refresh = true;
			}
			if (zoomout) {
				zoom(-0.1);
				refresh = true;
			}
			if (inciterations) {
				iterations += 5;
				refresh = true;
			}
			if (deciterations) {
				if (iterations > 5) {
					iterations -= 5;
					refresh = true;
				}
			}
			if (incwidth) {
				width += 10;
				refresh = true;
			}
			if (decwidth) {
				if (width > 10) {
					width -= 10;
					refresh = true;
				}
			}
			if (incheight) {
				height += 10;
				refresh = true;
			}
			if (decheight) {
				if (height > 10) {
					height -= 10;
					refresh = true;
				}
			}
		
			if (refresh)
				refresh();
		}
	}


	
	public static void main(String[] args) {
		PropertyManager.init();
		new MandelbrotViewer();
	}
}